// Copyright (c) 2020 by Robert Bosch GmbH. All rights reserved.
// Copyright (c) 2021 by Apex.AI Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// SPDX-License-Identifier: Apache-2.0

#include "iceoryx_posh/roudi/memory/memory_block.hpp"
#include "iceoryx_hoofs/cxx/attributes.hpp"

namespace iox
{
namespace roudi
{
void MemoryBlock::onMemoryAvailable(cxx::not_null<void*> memory IOX_MAYBE_UNUSED) noexcept
{
    // nothing to do in the default implementation
}

cxx::optional<void*> MemoryBlock::memory() const noexcept
{
    return m_memory ? cxx::make_optional<void*>(m_memory) : cxx::nullopt_t();
}

} // namespace roudi
} // namespace iox
